{
NOTICE:
This component was based on Borland's TDBImage code.

To do:
- Support for other formats in PasteFromClipBoard

}

{$DEFINE USE_JV_GIF} // Test
unit JvDBImage;

interface
uses
  Windows,
  Classes,
  Clipbrd,
  Controls,
  Db,
  DBCtrls,
  Forms,
  Graphics,
  Messages;

type
  // **********
  // TJvDBImage
  // **********

  TJvGetGtaphicClassEvent = procedure (Sender: TObject; Stream: TMemoryStream;
    var GraphicClass: TGraphicClass) of object;

  TJvDBImage = class(TDBImage)
  private
    FDataLink:TFieldDataLink;
    FOldPictureChange:TNotifyEvent;
//    FOldDataChange, FOldUpdateData:TNotifyEvent;
    FPictureLoaded:boolean;
    // ***** SESS: Do we need promote to protected ?
    // ***** SESS: New properties
    FOnGetGraphicClass: TJvGetGtaphicClassEvent;
    // ***** SESS: end new properties
    procedure NewPictureOnChange(Sender:TObject);
    procedure NewDataChange(Sender: TObject);
    procedure NewUpdateData(Sender: TObject);
    function GetAutoDisplay: boolean;
    procedure SetAutoDisplay(Value: boolean);
  protected
    procedure Paint; override;
    // **** SESS: New method(s)
    procedure CheckFieldType(AField: TField);
    function GetGraphic(AField: TField): TGraphic;
    procedure CreateHandle; override;

    // **** SESS: end new method(s)
  public
    procedure LoadPicture;
    // **** SESS: New method(s)
    class function GetGraphicClass(Stream: TStream): TGraphicClass;
    constructor Create(AOwner: TComponent);override;

    // **** SESS: end new method(s)
  published
    property AutoDisplay:boolean read GetAutoDisplay write SetAutoDisplay;
    // ***** SESS: New properties
    property OnGetGraphicClass: TJvGetGtaphicClassEvent read FOnGetGraphicClass
      write FOnGetGraphicClass;
    // ***** SESS: end new properties
  end;

  procedure RegisterGraphicSignature(ASignature: array of byte; AOffset: integer;
    AGraphicClass: TGraphicClass); overload;
  procedure RegisterGraphicSignature(ASignature: string; AOffset: integer;
    AGraphicClass: TGraphicClass); overload;

implementation
uses
  Contnrs,
  DBConsts,
  JPEG,
  {$IFDEF USE_JV_GIF}
  JvGIF,
  {$ENDIF}
  SysUtils;

// ***** SESS: Code to manage graphic's signatures. Should be public?

type
  TGraphicSignature = class
    Signature: string;
    Offset: integer;
    GraphicClass: TGraphicClass;
    constructor Create(ASignature: string; AOffset: integer;
      AGraphicClass: TGraphicClass);
    function IsThisSignature(Stream: TStream): boolean;
  end;

constructor TGraphicSignature.Create(ASignature: string; AOffset: integer;
  AGraphicClass: TGraphicClass);
begin
  inherited Create;
  Signature := ASignature;
  Offset := AOffset;
  GraphicClass := AGraphicClass;
end;

function TGraphicSignature.IsThisSignature(Stream: TStream): boolean;
var
  Buffer: string;
  Count: integer;
  BytesRead: integer;
begin
  Result := false;
  try
    Count := Length(Signature);
    SetLength(Buffer, Count);
    Stream.Position := Offset;
    BytesRead := Stream.Read(Buffer [1], Count);
    Result := (BytesRead = Count) and (Buffer = Signature);
  except
    // Ignore any error...
  end;
end;

var
  GraphicSignatures: TObjectList = nil;

procedure RegisterGraphicSignature(ASignature: string; AOffset: integer;
  AGraphicClass: TGraphicClass);
var
  GraphicSignature: TGraphicSignature;
begin
  // *DEV* - Avoid bad signatures
  GraphicSignature := TGraphicSignature.Create(ASignature, AOffset, AGraphicClass);
  try
    GraphicSignatures.Add(GraphicSignature)
  except
    GraphicSignature.Free;
  end;
end;

procedure RegisterGraphicSignature(ASignature: array of byte; AOffset: integer;
  AGraphicClass: TGraphicClass);
var
  Signature: string;
  i: integer;
begin
  SetLength(Signature, Length(ASignature));
  for i := Low(ASignature) to High(ASignature) do
    Signature [i + 1] := Char(ASignature[i]);
  RegisterGraphicSignature(Signature, AOffset, AGraphicClass);
end;

// **********
// TJvDBImage
// **********

constructor TJvDBImage.Create(AOwner: TComponent);
begin
  inherited;
  FOldPictureChange := Picture.OnChange;
  Picture.OnChange := NewPictureOnChange;
end;

procedure TJvDBImage.CreateHandle;
begin
  inherited;
  FDataLink := TFieldDataLink(SendMessage(Handle, CM_GETDATALINK, 0,0));
  if FDataLink <> nil then
  begin
    FOldDataChange := FDataLink.OnDataChange;
    FDataLink.OnDataChange := NewDataChange;
    FOldUpdateData := FDataLink.OnUpdateData;
    FDataLink.OnUpdateData := NewUpdateData;
  end;
end;

procedure TJvDBImage.NewPictureOnChange(Sender: TObject);
begin
  FOldPictureChange(Sender);
  FPictureLoaded := True;
end;

procedure TJvDBImage.NewDataChange(Sender: TObject);
begin
//  FOldDataChange(Sender);
  Picture.Graphic := nil;
  FPictureLoaded := False;
  if AutoDisplay then LoadPicture;
end;

procedure TJvDBImage.CheckFieldType(AField: TField);
begin
  with AField do
    if not IsBlob then
       DatabaseErrorFmt(SFieldTypeMismatch, [
         DisplayName,
         FieldTypeNames [ftBlob],
         FieldTypeNames [DataType]]);
end;

class function TJvDBImage.GetGraphicClass(Stream: TStream): TGraphicClass;
var
  i: integer;
  s: TGraphicSignature;
begin
  Result := nil;
  i := 0;
  while (Result = nil) and (i < GraphicSignatures.Count) do
  begin
    s := TGraphicSignature(GraphicSignatures [i]);
    if s.IsThisSignature(Stream) then
       Result := s.GraphicClass;
    Inc(i);
  end;
end;

function TJvDBImage.GetGraphic(AField: TField): TGraphic;
var
  GraphicClass: TGraphicClass;
  Stream: TMemoryStream;
begin
  CheckFieldType(AField);
  Result := nil;

  // If the field is null just exit
  if AField.IsNull then
     Exit;

  GraphicClass := nil;
  Stream := TMemoryStream.Create;
  try
    // Move blob data to Stream
    TBlobField(AField).SaveToStream(Stream);
    // Figure out which Graphic class is...
    GraphicClass := GetGraphicClass(Stream);
    // Call user event
    if Assigned(FOnGetGraphicClass) then
       FOnGetGraphicClass(Self, Stream, GraphicClass);
    // If we got one, load it..
    if GraphicClass <> nil then
    begin
      Result := GraphicClass.Create;
      try
        Stream.Position := 0;
        Result.LoadFromStream(Stream);
      except
        Result.Free;
        raise;
      end;
    end;
  finally
    Stream.Free;
  end;
end;

procedure TJvDBImage.Paint;
var
  Size: TSize;
  R: TRect;
  S: string;
  DrawPict: TPicture;
  Form: TCustomForm;
  Pal: HPalette;
begin
  if not Assigned(FDataLink) then Exit;
  with Canvas do
  begin
    Brush.Style := bsSolid;
    Brush.Color := Color;
    if FPictureLoaded or (csPaintCopy in ControlState) then
    begin
      DrawPict := TPicture.Create;
      Pal := 0;
      try
        if (csPaintCopy in ControlState) and
          Assigned(FDataLink.Field) and FDataLink.Field.IsBlob then
        begin
          // ***** SESS - Original code:
          // DrawPict.Assign(FDataLink.Field);
          // ***** SESS - New code
          DrawPict.Graphic := GetGraphic(FDataLink.Field);
          // ***** SESS - End new code
          if DrawPict.Graphic is TBitmap then
            DrawPict.Bitmap.IgnorePalette := QuickDraw;
        end
        else
        begin
          DrawPict.Assign(Picture);
          if Focused and (DrawPict.Graphic <> nil) and (DrawPict.Graphic.Palette <> 0) then
          begin { Control has focus, so realize the bitmap palette in foreground }
            Pal := SelectPalette(Handle, DrawPict.Graphic.Palette, False);
            RealizePalette(Handle);
          end;
        end;
        if Stretch then
          if (DrawPict.Graphic = nil) or DrawPict.Graphic.Empty then
            FillRect(ClientRect)
          else
            StretchDraw(ClientRect, DrawPict.Graphic)
        else
        begin
          SetRect(R, 0, 0, DrawPict.Width, DrawPict.Height);
          if Center then OffsetRect(R, (ClientWidth - DrawPict.Width) div 2,
            (ClientHeight - DrawPict.Height) div 2);
          StretchDraw(R, DrawPict.Graphic);
          ExcludeClipRect(Handle, R.Left, R.Top, R.Right, R.Bottom);
          FillRect(ClientRect);
          SelectClipRgn(Handle, 0);
        end;
      finally
        if Pal <> 0 then SelectPalette(Handle, Pal, True);
        DrawPict.Free;
      end;
    end
    else begin
      Font := Self.Font;
      if FDataLink.Field <> nil then
        S := FDataLink.Field.DisplayLabel
      else S := Name;
      S := '(' + S + ')';
      Size := TextExtent(S);
      R := ClientRect;
      TextRect(R, (R.Right - Size.cx) div 2, (R.Bottom - Size.cy) div 2, S);
    end;
    Form := GetParentForm(Self);
    if (Form <> nil) and (Form.ActiveControl = Self) and
      not (csDesigning in ComponentState) and
      not (csPaintCopy in ControlState) then
    begin
      Brush.Color := clWindowFrame;
      FrameRect(ClientRect);
    end;
  end;
end;

procedure TJvDBImage.LoadPicture;
begin
  try
    Picture.Graphic := GetGraphic(Field);
  except
    Picture.Graphic := nil;
    raise;
  end;
end;

procedure TJvDBImage.NewUpdateData(Sender: TObject);
var
  Stream: TMemoryStream;
begin
  CheckFieldType(Field);

  // If there is no graphic just clear field and exit
  if (Picture.Graphic = nil) then
  begin
    Field.Clear;
    Exit;
  end;
  Stream := TMemoryStream.Create;
  try
    Picture.Graphic.SaveToStream(Stream);
    Stream.Position := 0;
    TBlobField(Field).LoadFromStream(Stream);
  finally
    Stream.Free;
  end;
end;

// ***** SESS - Code to manage image signatures



function TJvDBImage.GetAutoDisplay: boolean;
begin
  Result := inherited AutoDisplay;
end;

procedure TJvDBImage.SetAutoDisplay(Value: boolean);
begin
  inherited AutoDisplay := Value;
  if Value then LoadPicture;
end;

initialization
  GraphicSignatures := TObjectList.Create(true);
  RegisterGraphicSignature('BM', 0, TBitMap);
  RegisterGraphicSignature([0, 0], 0, TIcon);
  RegisterGraphicSignature([$D7, $CD], 0, TMetaFile); // WMF
  RegisterGraphicSignature([0, 1], 0, TMetaFile);     // EMF
  RegisterGraphicSignature('JFIF', 6, TJPEGImage);
  {$IFDEF USE_JV_GIF}
  RegisterGraphicSignature('GIF', 0, TJvGIFImage);
  {$ENDIF}

finalization
  FreeAndNil(GraphicSignatures);

end.
